function [results, batch_info] = batch_run( id, varargin )

% [results, batch_info] = batch_run( id, X, iters, css, css_args,
%                                    comments, ['parallel', policy]
%                                    ['random_seed', seed] )
% [results, batch_info] = batch_run( id, ['parallel', policy] )
%
% id         Batch run id string or [] to run without using disk.
% X          dataset(s). May be a 1d cell array of data sets or a single data
%            set.
% iters      Number of iterations. Should be a single value or a vector of
%            equal length with css.
% css        Controller settings (cs) file handle(s). An IVGA run is made
%            for each given cs handle for each given data set as many times as
%            the 'iters' variable defines. The css argument may contain a
%            single handle or a cell array of handles.
% css_args   Optional arguments to the given controller settings
%            files. Syntax: 2-level cell array, so that css_args{cs}{n}
%            contains the n:th argument for cs entry 'cs', for example:
%              {{'runtime', 3600}, {'runtime', 600, 'name', 'short_run'}}
%            May also be [].
% comments   This string is stored into the results file. It could be a
%            good idea to specify the used data set(s) here, because no
%            information about them is stored automatically.
%
% ['parallel', policy]   Use parallel execution. The shell script
%                        launch_paraworker.sh can be used to launch the needed
%                        workers. They can be stopped by issuing the command
%                        'touch paraworkers/paraworkers_stop'. 'policy' defines
%                        the policies concerning system load and it has the
%                        following syntax:
%                          'minworkers=n, maxworkers=n, maxprocsrunning=n'
%                        with the first two having priority over the last one.
%                        See class 'paramaster' for details.
% ['random_seed', seed]  Use the given seed to generate random seeds for all
%                        jobs.
%
% The first call form creates a new batch run and starts executing it. An IVGA
% run is made for each given cs handle for each given data set as many times as
% the 'iters' variable defines. Individual runs are executed in random
% order. The results are saved into file 'batch_<id>_results.mat' in variable
% 'results', or just returned as the return value if no id was given.
%
% The results can be examined with the following functions:
%   ivga_show( <results> | <batch id>, [options] )
%   ivga_show_stats( <results> | <batch id>, [options] )
% For example:
%   batch_run( 'test1', ... )
%   ivga_show( 'test1' )
% or
%   results = batch_run( [], ... )
%   ivga_show( results )
%
% If a batch id was given, then intermediate results are stored to disk and all
% failures are (hopefully) tolerated, so that the second call form can be used
% to resume an interrupted batch run (execution is not resumed automatically in
% case of an error). If not parallel executing, then only result save failures
% are tolerated without interrupting the batch run (the failed operation(s)
% will be retried periodically).
% Warning: When resuming a parallel executed batch, the lock file in the
%          paraworkers directory must be manually deleted. Care should be taken
%          so that multiple batches do not use the same paraworkers directory
%          simultaneously, or the results will get mixed! It is ok to let the
%          workers run during a restart, but batch_run's behoviour is not
%          optimal in that case (some repeated work will be done).
%
% Status of a running batch can be seen from file batch_<id>_status.txt if an
% id was given.
%
% example:
%   load icann_1000x50 X
%   batch_run( 'merge_mod_3', X, 5, [@cs_adaptive @cs_adaptive_mmod3], ...
%              {{'max_split_depth', 2}, {}}, 'data set: icann_1000x50' )
%   ivga_show_stats( 'merge_mod_3' );


%  This file is a part of the Independent Variable Group Analysis package
% 
%  Copyright (C) 2001-2006 Esa Alhoniemi, Antti Honkela, Krista Lagus,
%  Jeremias Seppa, Harri Valpola, and Paul Wagner
% 
%  This program is free software; you can redistribute it and/or modify
%  it under the terms of the GNU General Public License as published by
%  the Free Software Foundation; either version 2, or (at your option)
%  any later version.
% 
%  This program is distributed in the hope that it will be useful,
%  but WITHOUT ANY WARRANTY; without even the implied warranty of
%  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%  GNU General Public License (included in file License.txt in the
%  program package) for more details.



% Implementation details
% ----------------------
%
% Controller settings entries (cs entries):
%
% Since revision 1.11, also an extra struct argument named 'batch_run_data' is
% passed to the cs functions. It contains the following entries:
%   .revision
%   .cs_ind
%   .dataset_ind
%   .iter
%
%
% The results file:
%
% The main results are saved into file batch_<id>_results.mat, which contains
% the following variables:
%   batch_info   : struct with fields:
%     .batch_id        : string or []
%     .comment         : string
%     .random_seed     : real
%     .timestamp_start : string
%     .timestamp_end   : string
%     .ivga_version    : string
%   results   : a 2d or 3d cell array of ivga() result structs. some
%               elements can be empty. dimensions:
%               dim 1 = cs entry (this dimension will be omitted and the
%                       rest shifted left by one if only one entry was
%                       provided)
%               dim 2 = data set
%               dim 3 = iteration

% $Id: batch_run.m,v 1.44 2006/04/14 17:29:41 pwagner Exp $
% $Name: ivga-1_14_3 $

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


% set constants
IVGA_VERSION  = '$Name: ivga-1_14_3 $';
IVGA_REVISION = '$Revision: 1.44 $';
STDOUT_REPLACE_MODE_ENABLED = ...
    environment( 'has_feature', 'addressable_cursor' );
STATE_VARS = { 'X', 'css', 'iters', 'comments', 'rseed', ...
               'jobs', 'done', 'saved', 'start_data' };




% -- initialize --


% detect mode ('start' or 'cont') and deal varargs
if nargin > 1 && ~ischar(varargin{1})
  % more args than just the batch id were given, and the first vararg is not
  % an isarg() argument name string --> form 1, start a new batch
  [X, iters, css, css_args, comments] = deal( varargin{1:5} );
  varargin(1:5) = [];
  [flg, rseed] = isarg('random_seed'); if ~flg, rseed = sum(100*clock); end
  use_disk = ~isempty(id);
  mode = 'start';
else
  % only the batch id was given, or the first vararg is an isarg() argument
  % name string --> form 2, continue an interrupted batch
  mode = 'cont';
  use_disk = true;
end


% check that id does not contain spaces
if ~isempty(strfind(id, ' ')), error( 'Id may not contain spaces!' ); end

% check for sleep feature if in parallel mode
[parallel, policy] = isarg('parallel');
if parallel && ~environment( 'has_feature', 'sleep' )
  error([ 'Current environment does not support the sleep feature, ' ...
          'which is required by the ''parallel'' mode!' ]);
end

% check the global quiet flag
global IVGA_CONFIG;
if isfield( IVGA_CONFIG, 'QUIET' ), QUIET = IVGA_CONFIG.QUIET;
else QUIET = false; end

% convert 'css' into a cell array if it is a function handle array
if exist( 'css', 'var' ) && isa(css, 'function_handle')
  if length(css) ~= 1
    % allow the old syntax with function handle vectors, but issue a warning
    warning([ 'The old syntax with the css argument being a function ' ...
              'handle vector is deprecated! A cell array should be ' ...
              'used instead.' ]);
    wstate = warning('off'); css = num2cell(css); warning(wstate);
  else
    css = {css};
  end
end


% check call form and initialize
switch mode
 case 'cont'
  
  % continue an interrupted batch: load the state
  if ~QUIET, fprintf( 'Continuing an interrupted batch:\n' ); end
  
  % try to lock the batch
  if ~QUIET, fprintf( '  Locking the batch..' ); end
  failed = 0;
  eval( 'batch_lock = lockfile( [], ''new'', [''batch_run_'' id] );', ...
        'failed = 1;' );
  if failed
    if QUIET || ~environment( 'has_feature', 'sleep' )
      fprintf( '\n' );
      fprintf('The batch is locked! Another process is already running it\n');
      fprintf('or was running it and crashed.\n');
      fprintf(['To continue, delete the file ' ...
               '''batch_run_%s_lockfile.mat''\n'], id);
      error('The batch is locked!');
    end
    fprintf( '\n' );
    fprintf( 'The batch is locked! Another process is already running it\n' );
    fprintf( 'or was running it and crashed.\n\n' );
    fprintf( 'PRESS CTRL-C IN 10 SECONDS to abort,\n' );
    fprintf( 'otherwise the batch ownership will be taken.\n\n' );
    fprintf( 'Waiting 10 seconds..' );
    errcode = system('sleep 10');
    if errcode, fprintf( 'aborted!\n' ); error('Batch locked!'); end
    batch_lock = lockfile( [], 'new', ['batch_run_' id], 'force' );
    fprintf( 'ownership taken.\n\n' );
  end
  if ~QUIET, fprintf( 'OK\n' ); end
  
  % load the state
  if ~QUIET, fprintf( '  Loading state from file..' ); end
  load([ 'batch_run_' id '_temp.mat' ]);
  done = saved;
  if ~QUIET, fprintf( 'OK\n\n\n' ); end
  
 case 'start'
  
  % start a new batch: create a new state
  if ~QUIET, fprintf( 'Starting a new batch:\n' ); end
  
  % check for error conditions
  if ~QUIET, fprintf( '  Checking for error conditions..' ); end
  if ~( isempty(id) || (ischar(id) && isvector_(id)) )
    error( '''id'' should be a string!' );
  end
  if ~isnumeric(iters), error( '''iters'' should be numeric!' ); end
  if ~( isscalar_(iters) || ...
        ( isvector_(iters) && length(iters) == length(css) ))
    error([ '''iters'' should be either a single value or a vector of ' ...
            'equal length with ''css''!' ]);
  end
  if ~iscell(css) || ~isa(css{1}, 'function_handle')
    error( '''css'' should be a cell array of function handles!' );
  end
  if ~( isempty(css_args) || (iscell(css_args) && iscell(css_args{1})) )
    error( '''css_args'' should either be empty or a 2-level cell array!' );
  end
  if ~isempty(css_args) && length(css_args) ~= length(css)
    error( 'Length of ''css_args'' should match the length of ''css''!' );
  end
  if strcmp( comments, 'parallel' )
    if ~QUIET, fprintf( '\n\n' ); end
    warning( '''comments'' seems to contain the keyword ''parallel''..' );
    if ~QUIET, fprintf( '\n\n' ); end
  end
  if ~use_disk && parallel
    error('A batch id must be given when using parallel execution!');
  end
  if use_disk && exist( ['batch_run_' id '_temp.mat'], 'file' )
    error([ 'An unfinished batch run with the same id exists! ' ...
            '(Delete the file "batch_run_' id '_temp.mat" ' ...
            'to overwrite it or use the ''continue'' option to continue ' ...
            'executing it)' ]);
  end
  if use_disk && exist( ['batch_' id '_results.mat'], 'file' )
    error([ 'A finished batch run with the same id exists! ' ...
            '(Delete the file "batch_' id '_results.mat" ' ...
            'to overwrite it)' ]);
  end
  if ~QUIET, fprintf( 'OK\n' ); end
  
  % store information about startup
  start_data.start_timestamp = datestr(now);
  
  % force lock the batch
  if use_disk
    if ~QUIET, fprintf( '  Locking the batch..' ); end
    batch_lock = lockfile( [], 'new', ['batch_run_' id], 'force' );
    if ~QUIET, fprintf( 'OK\n' ); end
  end
  
  % create the new state
  if ~QUIET, fprintf( '  Creating the state..' ); end
  if ~isa(X, 'cell'), X = {X}; end
  if length(iters) == 1, iters = repmat( iters, 1, length(css) ); end
  if isempty(css_args), css_args = repmat( {{}}, 1, length(css) ); end
  jobs  = create_jobs( X, iters, css, css_args, rseed );
  done  = logical(zeros(1,length(jobs)));
  saved = logical(zeros(1,length(jobs)));
  
  % save the state to disk (if use_disk is true)
  if use_disk
    if ~QUIET, fprintf( 'OK\n  Saving state to file..' ); end
    if environment('has_feature', 'working_copymovefile')
      safe_save( ['batch_run_' id '_temp.mat'], STATE_VARS );
    else
      save( ['batch_run_' id '_temp.mat'], STATE_VARS{:} );
    end
    if ~QUIET, fprintf( 'OK\n\n\n' ); end
  end
  
 otherwise
  
  error('Wrong number of arguments!');
  
end


% initialize the result storage
if use_disk
  rs_filename   = [ 'batch_run_' id '_result_storage.mat' ];
  rs_batch_lock = batch_lock;
else
  rs_filename   = [];
  rs_batch_lock = [];
end
result_storage = make_result_storage( rs_batch_lock, rs_filename, ...
                                      strcmp( mode, 'cont' ) );


% initialize paramaster and/or some related variables
if parallel
  para_master = paramaster( 'paraworkers/', 'policy', policy, 'quiet' );
  stats;   % introduce the class
else
  results = struct( 'job_id', [], 'result', [] );   % ???
end




% -- run the jobs --


% loop through all undone jobs
while any(~done)
  
  % fetch the current job struct and print some status information
  job_ind = find(~done);
  job_ind = job_ind(1);
  job = jobs(job_ind);
  
  % call IVGA and get results (use the internal call syntax of ivga())
  cs_args = { ...
      'batch_run_data', struct( ...
          'revision', IVGA_REVISION, ...
          'cs_ind', job.cs, ...
          'dataset_ind', job.dataset, ...
          'iter', job.iter ), ...
      job.args{:} };
  ivga_args = { X{job.dataset}, css{job.cs}, [], job.rseed, cs_args{:} };
  if parallel
    temp = feval( css{job.cs}, X{job.dataset}, [], cs_args{:} );
    write_status( id, [ '  Parallel executed with policy:\n    ' ...
                        strrep( evalc( 'type paraworkers/policy.txt' ), ...
                                char(10), '' ) '\n' ...
                        '  Next job to be enqueued: ' ...
                        num2str(length(find(done)) + 1) '/' ...
                        num2str(length(jobs)) '. ' ...
                        'Pending save operations: ' ...
                        num2str(length(find(done & ~saved))) '\n' ], ...
                  'replace_mode' );
    status = 0;
    while status == 0
      [para_master, status] = enqueue( para_master, ...
                                       { 'ivga', ivga_args{:} }, [], ...
                                       'block', 'quiet', 'job_id', job_ind );
      if status == 0
        warning('Problems while enqueueing the next job! Delaying..');
        system( 'sleep 60' ); i=1e6; while i, i=i-1; end
      end
    end
    results = get_results( para_master );
  else
    if ~QUIET, fprintf( '\n\n\n\n\n\n\n\n' ); end
    msg = [ '  job ' num2str(length(find(done)) + 1) '/' ...
                        num2str(length(jobs)) '.' ];
    if use_disk, msg = [ msg ...
                         ' Pending save operations: ' ...
                         num2str(length(find(done & ~saved))) '\n' ];
    end
    write_status( id, msg );
    if ~QUIET, fprintf( '========================================\n\n' ); end
    results.job_id = job_ind;
    results.result = ivga( ivga_args{:} );
  end
  
  % mark the just executed job as done
  done(job_ind) = true;
  
  % store results
  [result_storage, done, saved] = ...
      store_new_results( result_storage, done, saved, jobs, results );
  
  % save the current state
  if use_disk
    batch_lock = lockfile( batch_lock, 'check' );
    if environment('has_feature', 'working_copymovefile')
      safe_save( ['batch_run_' id '_temp.mat'], STATE_VARS, 'ignore_errors' );
    else
      try save( ['batch_run_' id '_temp.mat'], STATE_VARS{:} );
      catch; end
    end
  end
  
end


% If parallel executing, then wait until all results have come back from the
% workers AND saved successfully into the result storage (the latter just
% simplifies things a bit). If not parallel executing, then it is ok to proceed
% with unsaved results in the result storage.
if parallel
  while ~all(saved)
    
    % write status and sleep 1 min
    write_status( id, [ '  ' num2str(length(find(~saved))) ...
                        ' jobs still running. Waiting..\n' ], 'replace_mode' );
    system( 'sleep 60' ); i=1e6; while i, i=i-1; end
    
    % check for new results
    results = get_results( para_master );
    
    % store results
    [result_storage, done, saved] = ...
        store_new_results( result_storage, done, saved, jobs, results );
    
    % save the current state
    if use_disk   % redundant, should be always true in parallel mode
      batch_lock = lockfile( batch_lock, 'check' );
      if environment('has_feature', 'working_copymovefile')
        safe_save( ['batch_run_' id '_temp.mat'], STATE_VARS, 'ignore_errors');
      else
        try save( ['batch_run_' id '_temp.mat'], STATE_VARS{:} );
        catch; end
      end
    end
    
  end
end


% delete paramaster if parallel executing
if parallel, para_master = destruct( para_master ); end




% -- all jobs executed, collect the results --


% write status info: all done, post-processing
write_status( id, [ '  All jobs executed successfully. Combining the ' ...
                    'final result array..\n' ], ...
              'replace_mode' );

% retrieve all results into 'results'
results = retrieve_and_reshape_results( result_storage, jobs );

% sort iterations by total cost
for i=1:size(results, 1)
  for j=1:size(results, 2)
    inds = ~cellfun( 'isempty', results(i,j,:) );
    existing_results = cell2mat( results(i,j,inds) );
    [dummy, order] = sort( [existing_results.cost] );
    results(i,j,inds) = num2cell( existing_results(order) );
  end
end

% collect some general information about the batch
batch_info = struct( 'batch_id', id, ...
                     'comment', comments, ...
                     'random_seed', rseed, ...
                     'timestamp_start', start_data.start_timestamp, ...
                     'timestamp_end', datestr(now), ...
                     'ivga_version', IVGA_VERSION );

% remove the first dim if only one entry (this is shifted back in
% collect_from_results())
if length(css) == 1
  results = shiftdim( results, 1 );
else
  % this is needed if the results have multiple entries but only one
  % repetition (rare, but..)
  batch_info.multiple_entries = 'dummy_value';
end

if use_disk
  % create the final results file
  write_status( id, [ '  Writing the main results file..\n' ], ...
                'replace_mode' );
  save( ['batch_' id '_results.mat'], 'batch_info', 'results' );
  
  % clean up temporary files
  batch_lock = lockfile( batch_lock, 'check' );
  delete([ 'batch_run_' id '_temp.mat' ]);
  if ~isempty( rs_filename ), delete( rs_filename ); end
  batch_lock = lockfile( batch_lock, 'delete' );
end

% write status info: all done
if use_disk
  write_status( id, [ 'Batch run completed successfully!\n' ...
                      '  Results are in the file: ' ...
                      '"batch_' id '_results.mat"\n' ...
                      '  You can use the functions ivga_show() and ' ...
                      'ivga_show_stats() to work with it:\n' ...
                      '    ivga_show( ''' id ''', [options] )\n' ...
                      '    ivga_show_stats( ''' id ''', [options] )\n\n' ] );
  delete([ 'batch_' id '_status.txt' ]);
else
  write_status( id, [ 'Batch run completed successfully!\n' ...
                      '  You can use the functions ivga_show() and ' ...
                      'ivga_show_stats() to work with\n' ...
                      '  the returned results:\n' ...
                      '    ivga_show( <result>, [options] )\n' ...
                      '    ivga_show_stats( <result>, [options] )\n\n' ] );
end

return;








function jobs = create_jobs( X, iters, css, css_args, rseed )

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


rand('state', rseed);

n_datasets = length(X);
n_css      = length(css);

jobs = struct( 'cs', {}, 'dataset', {}, 'iter', {}, 'args', {}, ...
               'rseed', {} );
for i_cs = 1:n_css
  for i_dataset = 1:n_datasets
    jobs = [ jobs, ...
             struct( 'cs', i_cs, 'dataset', i_dataset, 'iter', ...
                     num2cell(1:iters(i_cs)), 'args', {css_args{i_cs}}, ...
                     'rseed', num2cell( 1e6 * rand(1,iters(i_cs)) ) ) ];
  end
end
jobs = jobs(randperm(length(jobs)));

return;




function write_status( id, status, varargin )

%

% 'id' must be either the name of the batch, or [] if running without using
% disk. Accepted options: 'replace_mode'.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


rows = 5;

if ~isempty(id)
  status = [ 'IVGA batch run "' id '":\n' status ];
else
  status = [ 'IVGA batch run:\n' status ];
end
status = [ status repmat( '\n', 1, rows - length(strfind(status, '\n')) ) ];

if ~evalin('caller', 'QUIET')
  if evalin('caller', 'STDOUT_REPLACE_MODE_ENABLED')
    fprintf( strrep( status, '\n', [27 '[K\n'] ) );
  else
    fprintf( status );
  end
  if evalin('caller', 'STDOUT_REPLACE_MODE_ENABLED') && isarg('replace_mode')
    fprintf( '%c[%dA', 27, rows );
  end
end

% write the status string to batch_<id>_status.txt
if ~isempty(id)
  fid = fopen( [ 'batch_' id '_status.txt' ], 'wt' );
  if fid ~= -1, fprintf( fid, status ); fclose( fid ); end
end

return;




function this = lockfile( this, method, varargin )

% this = lockfile( [], 'new', id )
% this = lockfile( [], 'delete' )
% this = lockfile( this, 'check' )
%
% Creates a lockfile on construction. Method 'check' checks if the created
% lockfile is still intact and generates an error if not. Destructor deletes
% the lockfile (if it is owned) and invalidates 'this'.
%
% Meant for long interval locking. Will leak if parallel executed.. 

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


switch method
  
 case 'new'
  clear this; id = varargin{1};
  
  % create 'this'
  this.filename = [ id '_lockfile.mat' ];
  this.code = num2str(rand * sum(100*clock), 32);
  
  % create lockfile
  if exist( this.filename, 'file' ) && ~isarg('force')
    error([ 'Lockfile ''' this.filename ''' already exists!' ]);
  end
  code = this.code;
  save( this.filename, 'code' );
  
  return;
  
 case 'delete'
  owned = 1; eval( 'this = lockfile( this, ''check'' );', 'owned = 0;' );
  if owned, delete( this.filename ); end
  this = [];
  return;
  
 case 'check'
  load( this.filename, 'code' );
  if length(code) ~= length(this.code) || any(code ~= this.code)
    error([ 'Lock ownership lost!' ]);
  end
  return;
  
end

return;




function [result_storage, done, saved] = ...
    store_new_results( result_storage, done, saved, jobs, results )

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


for result=results
  
  % fetch the job data
  job_ind = result.job_id;
  job     = jobs(job_ind);
  
  % mark it done (needed with restarted, parallel executed batches)
  done(job_ind) = true;
  
  % purge the result
  result.result = purge_state( result.result );
  
  % store the result
  result_storage = result_storage_add( result_storage, ...
                                       job_ind, result.result );
  
end

[result_storage, just_saved] = result_storage_flush( result_storage );
saved(just_saved) = true;

return;




function results = retrieve_and_reshape_results( result_storage, jobs )

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


% retrieve the stored results as a 1-dimensional cell array
results_1d = result_storage_get_contents( result_storage );

% check that all needed results are present
if length(results_1d) ~= length(jobs) || ...
      any( cellfun( 'isempty', results_1d ) )
  
  % inform the user
  fprintf( '\n\n\n\n\n\n\n\n' );
  fprintf( [ 'WARNING: An internal error occurred and some results are ' ...
             'missing!\n         The final result array will be ' ...
             'incomplete.\n\n' ] );
  warning('Internal error: Some results are missing!');
  
  % complete the result array with empty cells
  results_1d_new = repmat( {[]}, 1, length(jobs) );
  results_1d_new( ~cellfun( 'isempty', results_1d ) ) = ...
      results_1d( ~cellfun( 'isempty', results_1d ) );
  results_1d = results_1d_new;
  
end

% reshape the results
for i=1:length(results_1d)
  if ~isempty( results_1d{i} )
    results{ jobs(i).cs, jobs(i).dataset, jobs(i).iter } = results_1d{i};
  end
end

return;








% ---- result_storage 'methods' begin ----




function this = make_result_storage( batch_lock, disk_backup, ...
                                     use_existing_disk_backup )

% Constructs a new result storage 'object'.
%
% batch_lock                 An initialized batch lock 'object'. Can be [] if
%                            'disk_backup' is also empty.
% disk_backup                This can be either a filename to be used as a
%                            backup file, or it can be the empty matrix [] to
%                            disable backing up to disk.
% use_existing_disk_backup   Decides whether the contents of a possibly
%                            existing backup file with the same batch id
%                            should be used or not.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

this.batch_lock  = batch_lock;
this.disk_backup = disk_backup;
this.results     = {};

% if a disk backup filename was given, and the file does not exist or the
% existing one should not be used, then clear and initialize the file
if ~isempty(this.disk_backup) && ...
      ( ~exist( this.disk_backup, 'file' ) || ~use_existing_disk_backup )
  dummy = [];
  save( this.disk_backup, 'dummy' );
end

return;


function this = result_storage_add( this, job_ind, result )

% Adds the given result to the result storage. Nothing is written to disk at
% this point.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

this.results{job_ind} = result;

return;


function [this, just_saved] = result_storage_flush( this )

% Flushes all results still in memory to disk if a disk backup file is being
% used, otherwise do nothing. Job inds of successfully saved results are
% returned in 'just_saved' and the results are removed from memory.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

just_saved = [];

if ~isempty(this.disk_backup)
  lockfile( this.batch_lock, 'check' );
  for i=1:length(this.results)
    if ~isempty( this.results{i} )
      var_name = [ 'result_' num2str(i) ];
      failed = 0;
      if environment('has_feature', 'working_copymovefile')
        eval( 'safe_append( this.disk_backup, var_name, this.results{i} );',...
              'failed = 1;' );
      else
        eval([var_name ' = this.results{i};']);
        eval( 'save( this.disk_backup, var_name, ''-append'' );',...
              'failed = 1;' );
      end
      if ~failed, this.results{i} = []; just_saved = [ just_saved i ]; end
    end
  end
end

return;


function results = result_storage_get_contents( this )

% Returns all stored results as a cell array of results, with job inds
% corresponding to the indices of the array. All results stored to disk are
% read back at this point.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% load all results present in the disk backup file if one was used
if ~isempty(this.disk_backup)
  load( this.disk_backup );
  vars = who( 'result_*' );
  for i=1:length(vars)
    job_ind = sscanf( vars{i}, 'result_%d' );
    this.results{job_ind} = eval( vars{i} );
  end
end

results = this.results;

return;




% ---- result_storage 'methods' end ----
